#include "hierarchicalcontrol.h"

#include <map>
#include <vector>

#include "liftedstatebacktrackingsystem.h"
#include "additionaluserinterface.h"
template<class T> void addToVectorList(std::vector<std::pair<T, int> > &insegments, const T& firstargument, int value)
{
    auto it = std::find_if(insegments.begin(), insegments.end(), [&firstargument](const std::pair<T, int>& p) -> bool {return p.first == firstargument;});
    if (it == insegments.end()) {insegments.push_back(std::make_pair(firstargument, value)); return;};
    *it = std::make_pair(firstargument, it->second+value);
}

template<class T> int getVectorCount(std::vector<std::pair<T, int> > &insegments, const T& firstargument)
{
    auto it = std::find_if(insegments.begin(), insegments.end(), [&firstargument](const std::pair<T, int>& p) -> bool {return p.first == firstargument;});
    if (it == insegments.end()) return 0;
    return it->second;
}



RTSHierarchicalControl::RTSHierarchicalControl()
{
    /*struct TRTSEntityDesc {
        CUtiIDString m_name;
        int m_layer;
        std::vector<std::pair<CUtiIDString, int> > m_subentitycount;
        std::set<CUtiIDString> m_connectedstreets;
    };*/
    m_materials.push_back(IDSTR("wood"));
    m_materials.push_back(IDSTR("stone"));
    //m_materials.push_back(IDSTR("planks"));
    {
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_main_l2ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        d.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_mainbuildingID"), 1));
        d.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_building_woodchopperID"), 1));
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 2;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }
    {
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_mining_l2ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        d.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_building_woodchopperID"), 2));
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 2;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }
    {
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_streetAll_l2ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 2;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }
    {
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_streetAll_l1ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 1;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }
    {
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_streetAll_l1ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 1;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }{
        TRTSEntityDesc d;
        d.m_name = IDSTR("TEntity_city_l1ID");
        std::vector<std::pair<CUtiIDString, int> > subentitycount;
        d.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_main_l2ID"), 1));
       // d.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_mining_l2ID"), 3));
        std::set<CUtiIDString> connectedstreets;
        d.m_layer = 1;
        d.m_connectedstreets.insert(IDSTR("edgedir_x_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_x_down"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_up"));
        d.m_connectedstreets.insert(IDSTR("edgedir_z_down"));
        m_entitydesc.insert(std::make_pair(d.m_name, d));

    }
    {
    TRTSNPCDesc n;
    n.m_name =IDSTR("TEntity_npc_carrier_l2ID");

    n.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_npc_carrierID"), 2));
    n.m_layer = 2;
    m_npcdesc.insert(std::make_pair(n.m_name, n));
    //todo: add goal

    }
    {
    TRTSNPCDesc n;
    n.m_name =IDSTR("TEntity_npc_builder_l2ID");
    n.m_layer = 2;
    n.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_npc_builderID"), 2));
    m_npcdesc.insert(std::make_pair(n.m_name, n));

    }
    {
    TRTSNPCDesc n;
    n.m_name =IDSTR("TEntity_npc_carrier_l1ID");
    n.m_layer = 1;
    n.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_npc_carrier_l2ID"), 2));
    m_npcdesc.insert(std::make_pair(n.m_name, n));

    }

    {
    TRTSNPCDesc n;
    n.m_name =IDSTR("TEntity_npc_builder_l1ID");
    n.m_layer = 1;
    n.m_subentitycount.push_back(std::make_pair(IDSTR("TEntity_npc_builder_l2ID"), 2));
    m_npcdesc.insert(std::make_pair(n.m_name, n));

    }

    m_object_to_level.insert(std::make_pair(IDSTR("TNodeID3"), 3));
    m_object_to_level.insert(std::make_pair(IDSTR("TNodeID0"), 0));
    m_object_to_level.insert(std::make_pair(IDSTR("TNodeID1"), 1));
    m_object_to_level.insert(std::make_pair(IDSTR("TNodeID2"), 2));
    m_object_to_level.insert(std::make_pair(IDSTR("TEdgeID3"), 3));
    m_object_to_level.insert(std::make_pair(IDSTR("TEdgeID0"), 0));
    m_object_to_level.insert(std::make_pair(IDSTR("TEdgeID1"), 1));
    m_object_to_level.insert(std::make_pair(IDSTR("TEdgeID2"), 2));


    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_node"),3| SEEALWAYS));


    m_literal_name_to_level.insert(std::make_pair(IDSTR(""), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_node"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_parent"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_nodetype"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_blockedbynpc"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_usedbybuilding"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_enabled"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_edgedir_x_up"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_edgedir_x_down"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_edgedir_z_up"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("node_edgedir_z_down"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_edgenode"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("edgenode_startnode"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("edgenode_endnode"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("edgenode_edgetype"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("edgenode_edgelength"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("edgenode_enabled"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metanode0"), 0|SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_player"),0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_usedbybuilding"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_nodetype"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_enabled"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_edgedir_x_up"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_edgedir_x_down"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_edgedir_z_up"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_edgedir_z_down"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaedgenode0"), 0|SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_startnode"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_endnode"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_edgelength"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_hasstreet"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_sublockednode"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode0_enabled"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metanode1"), 1 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_parent"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_player"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_usedbybuilding"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_nodetype"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_enabled"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_edgedir_x_up"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_edgedir_x_down"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_edgedir_z_up"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_edgedir_z_down"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaedgenode1"), 1 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_startnode"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_endnode"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_edgelength"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_hasstreet"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_sublockednode"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode1_enabled"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metanode2"), 2 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_parent"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_player"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_usedbybuilding"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_nodetype"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_enabled"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_edgedir_x_up"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_edgedir_x_down"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_edgedir_z_up"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_edgedir_z_down"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaedgenode2"), 2 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_startnode"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_endnode"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_edgelength"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_hasstreet"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_sublockednode"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaedgenode2_enabled"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_entity"), 3 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_player"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_npcowner"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_position"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_blocked"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_education"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaentity0"), 0 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_player"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_position"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_blocked"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_education"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaentity1"), 1 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_player"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_npcowner"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_position"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_blocked"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_education"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("unused_metaentity2"), 2 | SEEALWAYS));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_player"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_npcowner"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_position"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_blocked"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_education"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("num_greater"), -1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("num_add"), -1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("isequal"), -1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("head_unallocated"), -1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("head_next"), -1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("num_sub"), -1));

    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_health"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_happiness"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_produktionauslastung"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_livingbuildings"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_commercialbuildings"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_industrialbuildings"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode0_effectivitat"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_health"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_happiness"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_produktionauslastung"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_livingbuildings"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_commercialbuildings"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_industrialbuildings"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode1_effectivitat"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_health"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_happiness"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_produktionauslastung"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_livingbuildings"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_commercialbuildings"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_industrialbuildings"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metanode2_effectivitat"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_health"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_happiness"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_fatique"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_health"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_happiness"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_fatique"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_innode"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_health"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_happiness"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_fatique"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_innode"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_health"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_happiness"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_fatique"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_innode"), 2));

    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity0_workingcitizens"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity1_workingcitizens"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("metaentity2_workingcitizens"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("entity_workingcitizens"), 3));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("inventory0"), 0));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("inventory1"), 1));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("inventory2"), 2));
    m_literal_name_to_level.insert(std::make_pair(IDSTR("inventory3"), 3));

    m_donotfollow.insert(IDSTR("nonode3"));
    m_donotfollow.insert(IDSTR("noedge3"));
    m_donotfollow.insert(IDSTR("nonode0"));
    m_donotfollow.insert(IDSTR("noedge0"));
    m_donotfollow.insert(IDSTR("nonode1"));
    m_donotfollow.insert(IDSTR("noedge1"));
    m_donotfollow.insert(IDSTR("nonode2"));
    m_donotfollow.insert(IDSTR("noedge2"));
    m_donotfollow.insert(IDSTR("noplayer"));
    m_donotfollow.insert(IDSTR("noentity0"));
    m_donotfollow.insert(IDSTR("noentity1"));
    m_donotfollow.insert(IDSTR("noentity2"));
    m_donotfollow.insert(IDSTR("noentity3"));

}

CUtiIDString nameOfNodesInDomain(int layer)
{
  switch (layer) {
     case 3:  return CUtiIDString(UTISTR("node"));
     default: return CUtiIDString(UTISTR("metanode")+layer);
  }
};

CUtiIDString nameOfEntityInDomain(int layer)
{
  switch (layer) {
     case 3:  return CUtiIDString(UTISTR("entity"));
     default: return CUtiIDString(UTISTR("metaentity")+layer);
  }
};

CUtiIDString nameOfNodesInDomain(int layer, const CUtiString& subproperty)
{
  switch (layer) {
     case 3:  return CUtiIDString(UTISTR("node")+subproperty);
     default: return CUtiIDString(UTISTR("metanode")+layer+subproperty);
  }
};

CUtiIDString nameOfEntityInDomain(int layer, const CUtiString& subproperty)
{
  switch (layer) {
     case 3:  return CUtiIDString(UTISTR("entity")+subproperty);
     default: return CUtiIDString(UTISTR("metaentity")+layer+subproperty);
  }
};
CUtiIDString nameOfEdgeInDomain(int layer)
{
  switch (layer) {
     case 3: return CUtiIDString(UTISTR("edgenode"));
     default: return CUtiIDString(UTISTR("metaedgenode")+layer);
  }
};

bool RTSHierarchicalControl::updateInventory(Planner::IndependentWorldState* worldstate, const CUtiIDString& id,  const std::vector<CUtiIDString> &children, int layer)
{

    CUtiIDString id_type = worldstate->getType(id);
      auto npcdesc = m_npcdesc.find(id_type);

      Array<float> res(m_materials.size());
      res.fill(0.0f);

      //(inventory3 ?arg1 - TEntityID3 ?arg2 - TMaterial)
      for (auto subid: children) {
          for (int i = 0; i < m_materials.size(); i++) {
             res[i] += worldstate->readInventoryFluentProperty(layer+1, subid, m_materials.at(i));
          }
      }
      bool changed = false;
      for (int i = 0; i < m_materials.size(); i++) {
          bool result = worldstate->writeInventoryFluentProperty(layer, id, m_materials.at(i), res[i]);
         changed = changed || result;
      }

      return changed;

}

void RTSHierarchicalControl::handleSubNPCS(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, std::set<CUtiIDString>& stickynodes, int layer)
{

    CUtiIDString id_type = worldstate->getType(id);
      auto npcdesc = m_npcdesc.find(id_type);
    ;
    std::vector<CUtiIDString> children = worldstate->getParamAll(Lisp::list(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")),IDSTR("?param"), id));
    int childcount = npcdesc->second.getTotalSubEntityCount();
    float health = 0.0f;//(float)children.size() / float(childcount);
    bool changed = false;

    float inprimarysegment = 0.0f;
    std::vector<std::pair<CUtiIDString, int> > insegments;
    float resthealth = 0.0f;
    for (auto subid: children) {
        CUtiIDString str = worldstate->getParam(Lisp::list(nameOfEntityInDomain(layer+1, UTISTR("_position")), subid, IDSTR("?param")));


        CUtiIDString selectednode_parent = worldstate->getParam(Lisp::list(IDSTR("node_parent"), str, IDSTR("?param")));

        addToVectorList(insegments, selectednode_parent, 1);

        float childhealth = worldstate->readFluentProperty(subid, nameOfEntityInDomain(layer+1, UTISTR("_health")));
        resthealth += childhealth;
    }
    resthealth /= float(childcount);
    health = resthealth;

    auto it =  std::max_element(insegments.begin(), insegments.end(), [](const std::pair<CUtiIDString, int> & a1, const std::pair<CUtiIDString, int> & a2)->bool {return a1.second < a2.second;});

    changed = changed || updateInventory(worldstate, id, children, layer);

    if (it != insegments.end()) {
        //We do nothing, if is not build, health = 0
        inprimarysegment = (float)it->second / (float)children.size();
        changed = changed || worldstate->writeProperty(id, nameOfEntityInDomain(layer, UTISTR("_position")), it->first);
    }
    changed = changed || worldstate->writeFluentProperty(id, nameOfEntityInDomain(layer, UTISTR("_health")), health);
    changed = changed || worldstate->writeFluentProperty(id, nameOfEntityInDomain(layer, UTISTR("_innode")), inprimarysegment);

    if (children.size() == 0) {
        //dead, kill it
        changed = true;
        worldstate->deleteClass(id, layer);
    }
    if (changed) {
        //stickynodes.insert(worldstate->readProperty(nameOfEntityInDomain(layer, UTISTR("_npcowner")), id));
        stickynodes.insert(worldstate->getParam(Lisp::list(nameOfEntityInDomain(layer, UTISTR("_npcowner")), id, IDSTR("?param"))));
    }
    //TODO: Update inventory
}


void RTSHierarchicalControl::handleSubBuildings(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, int layer, std::vector<std::pair<CUtiIDString, int> > subentitycount,std::set<CUtiIDString> connectedstreets, std::set<CUtiIDString>& stickynodes)
{

    int totalbuildingcount = 0;
    for (auto el : subentitycount) totalbuildingcount += std::get<1>(el);

    CUtiIDString entityposition = worldstate->readProperty(id, nameOfEntityInDomain(layer, UTISTR("_position")));

    float health = subentitycount.size() == 0 ? 1.0f :0.0f;

    std::vector<CUtiIDString> children;
    std::vector<std::pair<CUtiIDString, int> > currentcount_str;
    worldstate->collectAllObjectsWithProperty(nameOfEntityInDomain(layer, UTISTR("_npcowner")), id, children);
    float citizens = 0;
    for (auto el: children) {
        CUtiIDString type = worldstate->getType(el);
        int targetcount = getVectorCount(subentitycount, type);
        int currentcount = getVectorCount(currentcount_str, type);
        CUtiIDString childposition = worldstate->readProperty(el, nameOfEntityInDomain(layer+1, UTISTR("_position")));

        CUtiIDString childposition_parent = worldstate->readProperty(childposition, CUtiIDString(nameOfNodesInDomain(layer+1).str()+ UTISTR("_parent"))); //metanode2_parent, metanode1_parent, metanode0_parent


        float childhealth = worldstate->readFluentProperty(el, nameOfEntityInDomain(layer+1, UTISTR("_workingcitizens")));
        citizens = childhealth;

        if (currentcount < targetcount && childposition == childposition_parent) {
            addToVectorList(currentcount_str, type, 1);
            health += 1.0f/float(totalbuildingcount);
        }

    }

    float edgehealth = 0.0f;
    std::set<CUtiIDString> subnodes;
    for (auto s: connectedstreets) {
        CUtiIDString edgeid = worldstate->readProperty(id, CUtiIDString(nameOfNodesInDomain(layer).str()+ UTISTR("_")+s.str()));
        CUtiIDString connectednode = worldstate->readProperty(edgeid,CUtiIDString(nameOfEdgeInDomain(layer).str()+  UTISTR("_sublockednode")));
        if (connectednode == IDSTR("noedge").addInt(layer+1)) {

        } else {
         edgehealth += 1.0f/float(connectedstreets.size());
         subnodes.insert(connectednode);
        }
    }
    //check if nodes are connected?

    std::set<CUtiIDString> allstreetnodes;
    if (subnodes.size()) {
    allstreetnodes.insert(*subnodes.begin());
    //must be in the domain, ignore it for now
    }

    bool changed = false;

    changed = changed || updateInventory(worldstate, id, children, layer);


    changed = changed || worldstate->writeFluentProperty(id, nameOfEntityInDomain(layer, UTISTR("_health")) /*IDSTR("entity_health")*/, health*0.5f+edgehealth*0.5f);
    changed = changed || worldstate->writeFluentProperty(id, nameOfEntityInDomain(layer, UTISTR("_workingcitizens")) /*IDSTR("entity_health")*/, health*0.5f+edgehealth*0.5f);
    if (children.size() == 0) {
        //dead, kill it
        changed = true;
        worldstate->deleteClass(id, layer);
    }
    if (changed) {
        stickynodes.insert(worldstate->readProperty(nameOfEntityInDomain(layer+1, UTISTR("_npcowner"))/*IDSTR("entity_npcowner")*/, id));
    }
}

void RTSHierarchicalControl::updateNode(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, int neighbourhooddistance, std::set<CUtiIDString>& stickynodes)
{
    CUtiIDString id_type = worldstate->getType(id);


    //layer 3
    if (id_type == IDSTR("TEntity_npc_carrierID")) {

    } else if (id_type == IDSTR("TEntity_mainbuildingID")) {

    } else if (id_type == IDSTR("TEntity_npc_builderID")) {

    } else if (id_type == IDSTR("TEntity_building_woodchopperID")) {

    } else if (m_npcdesc.find(id_type) != m_npcdesc.end()) {
        auto it = m_npcdesc.find(id_type);

            handleSubNPCS(worldstate, id, stickynodes, it->second.m_layer);
        } else if (m_entitydesc.find(id_type) != m_entitydesc.end()) {
            auto it = m_entitydesc.find(id_type);

            handleSubBuildings(worldstate, id, it->second.m_layer, it->second.m_subentitycount, it->second.m_connectedstreets, stickynodes);

        }
}

void createGoalAction(CStrings& newactions, DynArray1D<Lisp::LObject>& goals,DynArray1D<Planner::SFluent>& fluentgoals,
std::vector<std::pair<CUtiIDString, CUtiIDString> > goalparams, Planner::Domain* domain, Planner::PlayerHierarchicalData* datares)
{
    newactions.add(UTISTR("(:action goalfinished_")+datares->m_playername.str());
    CUtiString params;
    for (auto g: goalparams)
     {
        params += g.first.str() + UTISTR(" - ")+g.second.str() +UTISTR(" ");
    }
    newactions.add(UTISTR("  :parameters (")+ params +UTISTR(")"));




    newactions.add(UTISTR("  :precondition (and "));
    for (auto g : goals) {
        newactions.add(UTISTR("   ")+g.print());
    }


    newactions.add(UTISTR("  )"));
    CUtiIDString goal(UTISTR("win ")+datares->m_playername);
    newactions.add(UTISTR("  :effect (and (")+goal.str()+UTISTR(")  )"));
    newactions.add(UTISTR(")"));
   // datares->m_newpredicates.push_back(Lisp::list(goal));

    Planner::Task* task = domain->addExternalPrimitivTask(CUtiIDString(UTISTR("goalfinished_")+datares->m_playername.str()));

    for(auto g: goalparams) {
        task->newVars1(g.first, g.second);
    }
    task->addEffect(Lisp::list(IDSTR("win"), datares->m_playername));
    for (auto g : goals) {
#ifndef NDEBUG
     // std::wcout << "New Goal: " << g.print() << std::endl;
#endif
      task->addPrecondition(g);
   }
    for (auto g: fluentgoals)
    {
        task->addPrecondition(g);
    }


    datares->resulttask = task;
    datares->goalsasliterals.push_back(new Planner::Literal(IDSTR("win"), datares->m_playername));
    std::wcout << "createGoalAction = " << task->toString() << std::endl;
}

CUtiIDString RTSHierarchicalControl::getMultiParamByType(Planner::Task* t, const std::map<CUtiIDString, CUtiIDString>& params, const std::vector<CUtiIDString> &types, const std::vector<CUtiIDString> &list)
{
    for (auto & el: list)
    {
        auto it = params.find(el);
        if (it != params.end()) {
            //std::wcout << "param = " << it->first.str() << ", " << it->second.str()<< " type " << t->variableToType(it->first).str() << std::endl;
            if (std::find(types.begin(), types.end(), t->variableToType(it->first)) != types.end()) {
                return params.at(el);
            }
        }
    }
    std::wcout << "param not found = " << convertToString(params) << std::endl;
    std::wcout << t->toString() << std::endl;
    assert(false);
    return CUtiIDString();
}

CUtiIDString RTSHierarchicalControl::getMultiParam(const std::map<CUtiIDString, CUtiIDString>& params, const std::vector<CUtiIDString> & list)
{

    for (auto & el: list)
    {
        if (params.find(el) != params.end()) {
            return params.at(el);
        }
    }
    assert(false);
    return CUtiIDString();
}

CIVector2 nameToPosition(const CUtiIDString& str)
{
    CUtiString rs = str.str();
   int p1 = rs.Pos(UTISTR("_"));
   int p2 = rs.Pos(p1+1,UTISTR("_"));
   int p3 = rs.Pos(p2+1,UTISTR("_"));

   int layer = rs(p1+2, (p2-(p1+2))).ToInt();
   if (layer == 0) layer = 3; //
   int xpos =layer != 3 ? rs(p2+1, (p3-(p2+1))).ToInt() : rs(p2+2, (p3-(p2+1))).ToInt();
   int ypos =layer != 3 ? rs(p3+1, 100).ToInt() : rs(p3+2, 100).ToInt();

//   assert(layer != 3);
  // assert(layer != 0);
 //  assert(ypos != 0);
   return CIVector2(xpos, ypos);
}

class CCreateMoveHeuristic : public Planner::LiftedStateHeuristicSystem
{
protected:
    CUtiIDString m_parentobject;
    CUtiIDString m_goalnode;
    CUtiIDString m_player;
    CUtiIDString m_win;
public:
    CUtiIDString m_positionattribute;
    CUtiIDString m_npcownerattribute;
    CUtiIDString m_parentnodeattribute;
    DynArray1D<CUtiIDString> m_childvalue;
    DynArray1D<CIVector2> m_goals;
    bool m_hasgoalsgenerated = false;
    virtual void createInfo(CStrings& output)
    {
        //createSubGoals();
        output.add(UTISTR("m_parentobject = ")+m_parentobject);
        output.add(UTISTR("m_goalnode = ")+m_goalnode);
        output.add(UTISTR("m_player = ")+m_player);
        output.add(UTISTR("m_win = ")+m_win);
        output.add(UTISTR("m_positionattribute = ")+m_positionattribute);
        output.add(UTISTR("m_npcownerattribute = ")+m_npcownerattribute);
        output.add(UTISTR("m_parentnodeattribute = ")+m_parentnodeattribute);
        output.add(UTISTR("m_childvalue = ")+convertToString(m_childvalue));
        output.add(UTISTR("m_goals = ")+convertToString(m_goals));
    }
    CCreateMoveHeuristic(const CUtiIDString& parentobject, const CUtiIDString& goalnode, const CUtiIDString& player) : LiftedStateHeuristicSystem(nullptr)
    {

        m_parentobject = parentobject;
        m_goalnode = goalnode;
        m_win = IDSTR("win");

    }

    void addChild(const CUtiIDString& str)
    {
        m_childvalue.push_back(str);
    }
    void createSubGoals()
    {
        m_goals.clear();
        //now find all goal targets in the layer (layer+1)
        //finde alle kindknoten in subdomain
        for (auto &el : m_system->m_storedliterals) {
            if (el.m_name == m_parentnodeattribute && el.m_parameters[1] == m_goalnode) {

                //das ist ein knoten
                CUtiIDString obj = el.m_parameters[0];
                CIVector2 objpos = nameToPosition(obj);
                if (m_goals.indexOf(objpos) == -1)
                m_goals.push_back(objpos);

            }
        }

        //std::wcout << "m_goals = " << convertToString(m_goals)  << " m_goalnode " << m_goalnode.str()<< std::endl;
    }
    int getCurrentHeuristic(int player)
    {
        if (m_hasgoalsgenerated == false)  {
            createSubGoals();
            m_hasgoalsgenerated = true;
        }
        //steps to goal
        bool won = false;
        int distance_summed = 0;
        for (auto &el : m_system->m_storedliterals) {
            if (el.m_name == m_positionattribute && m_childvalue.indexOf(el.m_parameters[0]) != -1) {


                //std::wcout << "found element = " << el.toString() << std::endl;
                CUtiIDString node = el.m_parameters[1];
                CIVector2 objpos = nameToPosition(node);
                int mindistance = 1000;
                for (auto & goal : m_goals)
                {
                    mindistance = umin(mindistance, goal.absDistanceTo(objpos));
                }
                distance_summed += mindistance;


            } else if (el.m_name == m_win && el.m_parameters[0] == m_player) {
                won = true;
            }
        }
        return distance_summed +(won ? 0 : 1);

    }
};

class CCreateMoveInventoryHeuristic : public Planner::LiftedStateHeuristicSystem
{
protected:
    CUtiIDString m_parentobject;
    CUtiIDString m_player;
    CUtiIDString m_win;
public:
    CUtiIDString m_positionattribute;
    CUtiIDString m_npcownerattribute;
    CUtiIDString m_inventoryfluent;
    CUtiIDString m_parentnodeattribute;
    DynArray1D<CUtiIDString > m_npcvalue; //npc
    DynArray1D<CUtiIDString> m_buildingvalue; //buildings
    DynArray1D<int> m_npcvalue_count; //npc
    DynArray1D<int> m_buildingvalue_count; //buildings
    DynArray1D<CIVector2> m_buildings;
    bool m_move_to_npcs=false;
    int m_goalamount = 0;
    CUtiIDString m_materialtype;
    bool m_hasgoalsgenerated = false;
    CCreateMoveInventoryHeuristic(const CUtiIDString& parentobject, const CUtiIDString& player) : LiftedStateHeuristicSystem(nullptr)
    {

        m_parentobject = parentobject;
        m_win = IDSTR("win");

    }

    void createInfo(CStrings& output)
    {
        output.add(UTISTR("m_parentobject = ")+m_parentobject);
        output.add(UTISTR("m_player = ")+m_player);
        output.add(UTISTR("m_win = ")+m_win);
        output.add(UTISTR("m_positionattribute = ")+m_positionattribute);
        output.add(UTISTR("m_npcownerattribute = ")+m_npcownerattribute);
        output.add(UTISTR("m_parentnodeattribute = ")+m_parentnodeattribute);
        output.add(UTISTR("m_inventoryfluent = ")+m_inventoryfluent);
        output.add(UTISTR("m_npcvalue = ")+convertToString(m_npcvalue));
        output.add(UTISTR("m_buildingvalue = ")+convertToString(m_buildingvalue));
        output.add(UTISTR("m_npcvalue_count = ")+convertToString(m_npcvalue_count));
        output.add(UTISTR("m_buildingvalue_count = ")+convertToString(m_buildingvalue_count));
        output.add(UTISTR("m_buildings = ")+convertToString(m_buildings));
        output.add(UTISTR("m_move_to_npcs = ")+convertToString(m_move_to_npcs));
        output.add(UTISTR("m_goalamount = ")+convertToString(m_goalamount));
        output.add(UTISTR("m_materialtype = ")+convertToString(m_materialtype));

    }

    void addNPCChild(const CUtiIDString& str)
    {
        m_npcvalue.push_back(str);
        m_npcvalue_count.push_back(0);
    }
    void addBuildingChild(const CUtiIDString& str)
    {
        m_buildingvalue.push_back(str);
        m_buildingvalue_count.push_back(0);
    }
    void createSubGoals()
    {
        //now find all goal targets in the layer (layer+1)
        //finde alle kindknoten in subdomain
        for (auto &el : m_system->m_storedliterals) {
            if (el.m_name == m_positionattribute && m_buildingvalue.indexOf(el.m_parameters[0])) {

                //das ist ein knoten
                CUtiIDString obj = el.m_parameters[0];
                CIVector2 objpos = nameToPosition(obj);
                m_buildings.setAndCreateElement(m_buildingvalue.indexOf(el.m_parameters[0]), objpos);

            }
        }
       // safasd
        assert(false);
    }
    int getCurrentHeuristic(int player)
    {
        if (m_hasgoalsgenerated == false)  {
            createSubGoals();
            m_hasgoalsgenerated = true;
        }
        //steps to goal
        bool won = false;
        int distance_summed = 0;
        int totalmaterialcollected=0;

        for (int i = 0; i < m_system->m_fluentsliteral.size(); i++) {

            auto& el = m_system->m_fluentsliteral[i];

            int buildingindex = m_buildingvalue.indexOf(el.m_parameters[0]);
            int npcindex = m_npcvalue.indexOf(el.m_parameters[0]);
            if (el.m_name == m_inventoryfluent && buildingindex != -1 && el.m_parameters[1] == m_materialtype) {
                m_buildingvalue_count[buildingindex] = binround(m_system->m_fluents[i].m_value.first);
                if (m_move_to_npcs) {
                    totalmaterialcollected += m_buildingvalue_count[npcindex];
                }
            }
            if (el.m_name == m_inventoryfluent && npcindex != -1 && el.m_parameters[1] == m_materialtype) {

                m_npcvalue_count[npcindex] = binround(m_system->m_fluents[i].m_value.first);
                if (m_move_to_npcs) {
                    totalmaterialcollected += m_npcvalue_count[npcindex];
                }
            }
        }

        for (auto &el : m_system->m_storedliterals) {
            if (el.m_name == m_positionattribute && m_npcvalue.indexOf(el.m_parameters[0])) {


                CUtiIDString node = el.m_parameters[1];
                CIVector2 objpos = nameToPosition(node);
                int mindistance = 1000;
                for (int i = 0; i< m_buildings.size(); i++)
                {
                    if (m_buildingvalue_count[i] > 0) {
                    mindistance = umin(mindistance, m_buildings[i].absDistanceTo(objpos));
                    }
                }
                distance_summed += mindistance;


            } else if (el.m_name == m_win && el.m_parameters[0] == m_player) {
                won = true;
            }
        }
        if (m_goalamount-totalmaterialcollected <= 0) {
            distance_summed=0;
        }

        return distance_summed +(won ? 0 : 1)+ umax(0, m_goalamount-totalmaterialcollected);

    }
};
class CCreateGoalHeuristic : public Planner::LiftedStateHeuristicSystem
{
protected:
    CUtiIDString m_parentobject;
    CUtiIDString m_parentnode;
    CUtiIDString m_player;
    CUtiIDString m_win;
public:
    CUtiIDString m_positionattribute;
    CUtiIDString m_npcownerattribute;
    CUtiIDString m_parentnodeattribute;
    DynArray1D<CUtiIDString> m_typevalue;
    DynArray1D<bool> m_hastype;
    void addGoal(const CUtiIDString& str)
    {
        m_typevalue.push_back(str);
        m_hastype.push_back(false);
    }

    CCreateGoalHeuristic(const CUtiIDString& parentobject, const CUtiIDString& parentnode, const CUtiIDString& player) : LiftedStateHeuristicSystem(nullptr)
    {

        m_parentobject = parentobject;
        m_parentnode = parentnode;
        m_win = IDSTR("win");
    }


    void createInfo(CStrings& output)
    {
        output.add(UTISTR("m_parentobject = ")+m_parentobject);
        output.add(UTISTR("m_player = ")+m_player);
        output.add(UTISTR("m_win = ")+m_win);
        output.add(UTISTR("m_positionattribute = ")+m_positionattribute);
        output.add(UTISTR("m_npcownerattribute = ")+m_npcownerattribute);
        output.add(UTISTR("m_parentnodeattribute = ")+m_parentnodeattribute);
        output.add(UTISTR("m_typevalue = ")+convertToString(m_typevalue));
        output.add(UTISTR("m_hastype = ")+convertToString(m_hastype));
    }

    int getCurrentHeuristic(int player)
    {
        //steps to goal
        m_hastype.fill(false);
        bool won = false;
        //std::wcout << "Last one: " <<m_system->m_storedliterals[m_system->m_storedliterals.size()-1].toString()<< "/"<< m_win.str() << std::endl;

        for (int i = 0; i <  m_system->m_storedliterals.size(); i++) {
            auto& el = m_system->m_storedliterals[i];
            if (i == m_system->m_storedliterals.size()-1) {
               // std::wcout <<"found it" << std::endl;
            }
            if (el.m_name == m_npcownerattribute && el.m_parameters[1] == m_parentobject) {

                CUtiIDString obj = el.m_parameters[0];
                CUtiIDString type = m_system->m_problem->getObjectType(obj);

                for (int i = 0; i < m_typevalue.size(); i++) {
                    if (m_typevalue[i] == type && m_hastype[i] == false) {
                        m_hastype[i] = true;
                        break;
                    }
                }
            } else if (el.m_name == m_win/* && el.m_parameters[0] == m_player*/) {
                won = true;
            }
        }
        int notfulfilled = 0;
        for (auto el : m_hastype) {if (!el) notfulfilled++;}
        if (won) {
           // std::wcout << "won!" << std::endl;
        }
        return notfulfilled +(won ? 0 : 1);

    }
};

std::tuple<bool, CUtiIDString, Planner::LiftedStateHeuristicSystem*> RTSHierarchicalControl::createFromEffectsGoal(Planner::Task* t, const CUtiIDString& taskname, Planner::IndependentWorldState* worldstate, const std::map<CUtiIDString, CUtiIDString>& params,  const CUtiIDString& actionname, const CReferenceList<Planner::Literal>& effects, CStrings& goals, CStrings& newactions, CReferenceList<Planner::PlayerHierarchicalData> &hierarchicaldata)
{
    // goals.add(UTISTR("(finished)"));
    std::wcout << "TaskName: " << taskname.str() << std::endl;


     //TODO: wann brauchen wir adversairy?
     //Falls wir einen anderen Player haben im Feld der beeinträchtigt wird
     //Oder wir uberlegen uns auf der oberen Ebene, was der andere macht und kompinieren es

     uti_refptr<Planner::PlayerHierarchicalData> datares = new Planner::PlayerHierarchicalData();

     hierarchicaldata.push_back(datares);

     CUtiIDString player = getMultiParam(params, {IDSTR("?player")});
     datares->m_playername = player;
     datares->m_playeratom = Planner::StaticLiteral(IDSTR("activeplayer"), player);

   //  assert(params.find(IDSTR("?entity0_id"))!= params.end());

     CUtiIDString id;

     //todo: mark it as create entity!adsfsdfs

     id = getMultiParam(params, {/*IDSTR("?entity0_id"), IDSTR("?metaentity20_id"), IDSTR("?metaentity10_id"), IDSTR("?metaentity11_id"), IDSTR("?metaentity21_id")*/IDSTR("?operator")});
     assert(id.empty() == false);

     CUtiIDString id_type = worldstate->getType(id);
     Planner::Task* task = worldstate->getActionByName(taskname);
     assert(task != nullptr);
     std::wcout << std::endl;
     std::wcout <<"createFromEffectsGoal"<< std::endl;
     std::wcout << task->toString() << " with params" << convertToString(params) << std::endl;

     std::wcout << std::endl;
     assert(task->m_abstract);
     int currentlevel = 0;
     if (m_npcdesc.find(id_type) != m_npcdesc.end()) {
          auto & desc = m_npcdesc.find(id_type)->second;
          currentlevel = desc.m_layer;
     } else if (m_entitydesc.find(id_type) != m_entitydesc.end()) {
         auto & desc = m_entitydesc.find(id_type)->second;
         currentlevel = desc.m_layer;

     } else {
         assert(false);
     }

     if (actionname.str().endsWith(UTISTR("_create"))) {

         std::vector<std::pair<CUtiIDString, int> > goalobjects;
         int layer = -1;

         bool isnpcbuilding = false;
         CUtiIDString newobjectid = getMultiParam(params, {IDSTR("?newobject")});
         assert(newobjectid.empty() == false);
         CUtiIDString newobjectid_type = worldstate->getType(newobjectid);

         if (m_npcdesc.find(newobjectid_type) != m_npcdesc.end()) {
              auto & desc = m_npcdesc.find(newobjectid_type)->second;
              std::copy(desc.m_subentitycount.begin(), desc.m_subentitycount.end(), std::back_inserter(goalobjects));
              layer = desc.m_layer;
         } else if (m_entitydesc.find(newobjectid_type) != m_entitydesc.end()) {
             auto & desc = m_entitydesc.find(newobjectid_type)->second;
             std::copy(desc.m_subentitycount.begin(), desc.m_subentitycount.end(), std::back_inserter(goalobjects));
             layer = desc.m_layer;
             isnpcbuilding = true;

         } else {
             assert(false);
         }
       //  CUtiIDString builderid = getMultiParam(params, {IDSTR("?entity0_id"), IDSTR("?metaentity22_id")});

         DynArray1D<Lisp::LObject> goals;
         std::vector<std::pair<CUtiIDString, CUtiIDString> > goalparams;

         CUtiIDString targetnode = getMultiParam(params, {IDSTR("?noden0_n0id")});// worldstate->readProperty(id, nameOfEntityInDomain(layer, UTISTR("_position")));
         int index = 0;
         CUtiIDString prevobjid;
         CUtiIDString builderid;
         /*if (actionname == IDSTR("TEntity_main_l2ID_create") || actionname == IDSTR("TEntity_city_l1ID_create") || actionname == IDSTR("TEntity_mainbuildingID_create")
                 || ) {
         builderid= getMultiParamByType(t, params,{IDSTR("TEntity_npc_builder_l2ID"), IDSTR("TEntity_npc_builder_l1ID"), IDSTR("TEntity_npc_builderID")}, {IDSTR("?entity0_id"), IDSTR("?metaentity20_id"), IDSTR("?metaentity10_id"), IDSTR("?metaentity11_id"), IDSTR("?metaentity21_id"), IDSTR("?metaentity12_id"), IDSTR("?metaentity22_id")});
         } else {
                 builderid= getMultiParam(params, {IDSTR("?metaentity11_id"), IDSTR("?entity1_id")});

                 isnpcbuilding = true;
         }*/

         builderid = id;

         NeuralPainter_Build* move = new NeuralPainter_Build(layer+1);
         move->m_isnpcbuild = isnpcbuilding;
         move->m_typestobuild = goalobjects;
         worldstate->collectAllObjectsWithProperty(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")), builderid, move->m_builderobjects);
         if (move->m_builderobjects.size() == 0) {
             std::wcout << "Buildercount = 0, builderid = " << builderid.str() << " params = " << convertToString(params) << std::endl;
         }
         assert(move->m_builderobjects.size() != 0);//could be 0, but for now we see it as a bug

         for (auto el: goalobjects) {

             prevobjid = CUtiIDString();
             for (int i = 0; i < el.second; i++)
             {

                 CUtiIDString posvar =  IDSTR("?pos").addInt(index);
                 CUtiIDString newobjid =  IDSTR("?newobjid").addInt(index);

                 goals.push_back(Lisp::list(nameOfEntityInDomain(layer+1, UTISTR("_position")),newobjid, posvar));
                 goals.push_back(Lisp::list(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")), newobjid, newobjectid));

                 goalparams.push_back(std::make_pair(newobjid, el.first));
                 goalparams.push_back(std::make_pair(posvar, CUtiIDString(UTISTR("TNodeID")+(layer+1))));
                 goals.push_back(Lisp::list(CUtiIDString(nameOfNodesInDomain(layer+1).str()+UTISTR("_parent")), posvar, targetnode));

                 if (prevobjid.empty() == false) {

                     goals.push_back(Lisp::list(IDSTR("not"), Lisp::list(IDSTR("isequal"), newobjid, prevobjid)));
                 }
                 prevobjid = newobjid;
                 index++;
             }
         }

         DynArray1D<Planner::SFluent> fl;

         createGoalAction(newactions, goals, fl, goalparams, worldstate->getProblem()->m_domain.get(), datares.get());

         CCreateGoalHeuristic* cgh = new CCreateGoalHeuristic(newobjectid/*parentobject*/, targetnode /*parentknoten*/, player);
         cgh->m_positionattribute =CUtiIDString(nameOfEntityInDomain(layer+1, UTISTR("_position")));
         cgh->m_npcownerattribute =CUtiIDString(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")));
         cgh->m_parentnodeattribute =CUtiIDString(nameOfNodesInDomain(layer+1, UTISTR("_parent")));

         cgh->m_neuralpainter = move;



         for (auto el: goalobjects) {

             for (int i = 0; i < el.second; i++)
             {
                cgh->addGoal(el.first);
             }
         }
         return std::make_tuple(true, id, cgh);
     } else if (actionname.str().endsWith(UTISTR("_inventory_move_to_npc")) || actionname.str().endsWith(UTISTR("_inventory_npc_to_move"))) {
         int layer = -1;

         DynArray1D<Lisp::LObject> goals;
         DynArray1D<Planner::SFluent> ftgoal;
         std::vector<std::pair<CUtiIDString, CUtiIDString> > goalparams;
         if (m_npcdesc.find(id_type) != m_npcdesc.end()) {

             auto & desc = m_npcdesc.find(id_type)->second;
             layer = desc.m_layer;
         } else if (m_entitydesc.find(id_type) != m_entitydesc.end()){

             auto & desc = m_entitydesc.find(id_type)->second;
             layer = desc.m_layer;
         } else {
             assert(false);
         }
         assert(layer == currentlevel);



         CUtiIDString idgoal;

         idgoal = getMultiParam(params, { IDSTR("?goal")/*, IDSTR("?metanode21_usedbybuilding"), IDSTR("?metanode11_usedbybuilding")*/});
         assert(idgoal.empty() == false);

         std::vector<CUtiIDString> children;
         //CUtiIDString targetnode = params.at(IDSTR("?node2_id"));

         worldstate->collectAllObjectsWithProperty(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")), id, children);

         std::vector<CUtiIDString> children_goal;
         //CUtiIDString targetnode = params.at(IDSTR("?node2_id"));
         worldstate->collectAllObjectsWithProperty(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")), idgoal, children_goal);


         int tcount = 0;
         CUtiIDString tobjtype;
         for (auto & fluenteffect: task->m_effect_fluents) {
             Planner::Fluent_Value* firstparam = check_cast<Planner::Fluent_Value*>(fluenteffect.first.m_data.get());
             Planner::FluentTerm* secondparam = check_cast<Planner::FluentTerm*>(fluenteffect.second.m_data.get());
             Planner::Fluent_Constant* c  =  check_cast<Planner::Fluent_Constant*> (secondparam->m_exp2.get());
             if (firstparam->m_number->m_name == IDSTR("inventory").addInt(layer)) {
                // TODO: das sind variablen im firstparam. Ersetze sie durch parameter
                // TODO: Ziel kann auch das andere object sein
                 if (!(params.at(firstparam->m_number->m_arguments[0].strid()) == id)) {
                     //std::wcout << "Wrong id in move_to_inv: "  <<params.at(firstparam->m_number->m_arguments[0].strid()).str() << " != "
                     //           << id.str() <<" for task " << taskname.str() << std::endl;
                     continue;
                 }

                 assert(params.at(firstparam->m_number->m_arguments[0].strid()) == id);
                 CUtiIDString objecttype = params.at(firstparam->m_number->m_arguments[1].strid());
                 //gehe alle children durch und berechne goal menge
                 float allobjects = 0.0f;
                 DynArray1D<Lisp::LObject> sum;
                 for (auto ch: children) {
                     double objectsinchild = worldstate->readInventoryFluentProperty(layer+1, ch, objecttype);
                     allobjects += objectsinchild;
                     sum.push_back(Lisp::list(IDSTR("inventory").addInt(layer+1), ch, objecttype));
                 }
                 tobjtype = objecttype;
                 float targetobjectcount = allobjects + c->m_number;
                 tcount = targetobjectcount;
                 if (sum.size() == 0) continue;
                 Lisp::LObject currentelement = sum.at(0);
                 Planner::SFluent totalsum(sum[0]);
                 for (int i = 1; i < sum.size(); i++)
                 {
                     totalsum = Planner::SFluent(totalsum, Planner::gttAdd, Planner::SFluent(sum[i]));
                     currentelement = Lisp::list(IDSTR("+"), currentelement, sum[i]);
                 }
                 currentelement = Lisp::list(IDSTR(">"), currentelement, targetobjectcount);
                // goals.push_back(currentelement);
                 ftgoal.push_back(Planner::SFluent(totalsum, Planner::gttGreater, Planner::SFluent(targetobjectcount)));
             }
         }
         //Precise genug?

         assert(tobjtype.empty() == false);
         createGoalAction(newactions, goals, ftgoal, goalparams, worldstate->getProblem()->m_domain.get(),datares.get());

         CCreateMoveInventoryHeuristic* cgh = new CCreateMoveInventoryHeuristic(id/*parentobject*/, player);
         cgh->m_positionattribute =CUtiIDString(nameOfEntityInDomain(layer+1, UTISTR("_position")));
         cgh->m_npcownerattribute =CUtiIDString(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")));
         cgh->m_parentnodeattribute =CUtiIDString(nameOfNodesInDomain(layer+1, UTISTR("_parent")));
         cgh->m_inventoryfluent = IDSTR("inventory").addInt(layer+1);
         cgh->m_move_to_npcs = actionname.str().endsWith(UTISTR("_inventory_move_to_npc"));
         cgh->m_goalamount = tcount;
         cgh->m_materialtype = tobjtype;
         NeuralPainter_MoveInventory* move = new NeuralPainter_MoveInventory(nullptr, currentlevel+1);
         move->m_material = tobjtype;
         move->m_tonpc = cgh->m_move_to_npcs;


         CUtiIDString m_material;
         std::vector<CUtiIDString> m_carriersobjects;
         bool m_tonpc=true;



         cgh->m_neuralpainter = move;

         if (actionname.str().endsWith(UTISTR("_inventory_move_to_npc"))) {
             for (auto id3: children) {
                 cgh->addBuildingChild(id3);
                 std::wcout << "Building = " << id3.str() << std::endl;
             }
             for (auto id3: children_goal) {
                 cgh->addNPCChild(id3);
                 std::wcout << "NPC = " << id3.str() << std::endl;
                 move->m_carriersobjects.push_back(id3);
             }

         } else {

             for (auto id3: children) {
                 cgh->addNPCChild(id3);
                 std::wcout << "NPC = " << id3.str() << std::endl;
                 move->m_carriersobjects.push_back(id3);
             }
             for (auto id3: children_goal) {
                 cgh->addBuildingChild(id3);
                 std::wcout << "Building = " << id3.str() << std::endl;
             }
         }



         return std::make_tuple(true, id, cgh);
     } else if (actionname.str().endsWith(UTISTR("_move"))) {
         CUtiIDString targetnode;
         std::vector<CUtiIDString> children;

         NeuralPainter_Move* move = new NeuralPainter_Move(currentlevel+1);

         //std::vector<CUtiIDString> m_moveableobjects;
         //CUtiIDString m_parentgoalnode;


         int layer=0;
        if (m_npcdesc.find(id_type) != m_npcdesc.end()) {
            auto & desc = m_npcdesc.find(id_type)->second;
            layer = desc.m_layer;
            assert(layer == currentlevel);
            targetnode = getMultiParam(params, {IDSTR("?targetnode")/*, IDSTR("?metanode22_id"), IDSTR("?metanode12_id")*/});
            worldstate->collectAllObjectsWithProperty(nameOfEntityInDomain(layer+1, UTISTR("_npcowner")), id, children);
            assert(children.size() != 0);//could be 0, but for now we see it as a bug

            move->m_parentgoalnode = targetnode;
            std::vector<std::pair<CUtiIDString, int> > insegments;
            DynArray1D<Lisp::LObject> goals;
            std::vector<std::pair<CUtiIDString, CUtiIDString> > goalparams;
            int index = 0;
            for (auto id: children) {
                //CUtiIDString str = worldstate->readProperty(id, IDSTR("entity_position"));
                CUtiIDString posvar =  IDSTR("?pos").addInt(index);
                goals.push_back(Lisp::list(nameOfEntityInDomain(desc.m_layer+1, UTISTR("_position")), id, posvar));
                goalparams.push_back(std::make_pair(posvar,  CUtiIDString(UTISTR("TNodeID")+(layer+1))));
                goals.push_back(Lisp::list(CUtiIDString(nameOfNodesInDomain(desc.m_layer+1).str()+UTISTR("_parent")), posvar, targetnode));
                move->m_moveableobjects.push_back(id);

                index++;
               // addToVectorList(insegments, str, 1);
            }
            DynArray1D<Planner::SFluent> ft;
            createGoalAction(newactions, goals, ft, goalparams, worldstate->getProblem()->m_domain.get(), datares.get());
        } else {
            assert(false);
        }
        CCreateMoveHeuristic* cgh = new CCreateMoveHeuristic(id/*parentobject*/, targetnode /*parentknoten*/, player);
        cgh->m_positionattribute =CUtiIDString(nameOfEntityInDomain(currentlevel+1, UTISTR("_position")));
        cgh->m_npcownerattribute =CUtiIDString(nameOfEntityInDomain(currentlevel+1, UTISTR("_npcowner")));
        cgh->m_parentnodeattribute =CUtiIDString(nameOfNodesInDomain(currentlevel+1, UTISTR("_parent")));
        cgh->m_neuralpainter = move;
        for (auto id3: children) {
            cgh->addChild(id3);
        }

        return std::make_tuple(true, id, cgh);
    } else if (actionname.str().endsWith(UTISTR("_destroy"))) {

         return std::make_tuple(true, CUtiIDString(), nullptr);
     }



     return std::make_tuple(false, CUtiIDString(), nullptr);

}

int RTSHierarchicalControl::getLevelFromObjectSorts(const CUtiIDString& type)
{


    auto it = m_object_to_level.find(type);
    if (it != m_object_to_level.end()){
        return it->second;
    }


    assert(false);
    return -1;
}

int RTSHierarchicalControl::getLevelFromLiteral(const Planner::Literal* l)
{
    auto it = m_literal_name_to_level.find(l->m_name);
    if (it != m_literal_name_to_level.end()){
        return it->second;
    }

    return 0;
}

void RTSHierarchicalControl::createMainProblem(Planner::IndependentWorldState* worldstate, int level, Planner::IndependentWorldState* newworldstate)
{


    std::set<CUtiIDString> types;
    std::set<CUtiIDString> currentatomobjects;
    std::set<CUtiIDString> newatomobjects;
    std::set<const Planner::Literal*> alreadyadded;
    CUtiIDString id_str =IDSTR("ID");
    std::set<CUtiIDString> atomtypes;


    //TODO: Unused objects predicates? -> unused objects -> function isLevelIndependent
    for (int step = 0; step < 1; step++)  {
        newatomobjects = currentatomobjects;
        for (int i = 0; i < worldstate->getStoredLiteralCount(); i++) {
           const Planner::Literal* l = &worldstate->getStoredLiteral(i);
           int literallevel = getLevelFromLiteral(l);
           if ((literallevel & SEEALWAYS) != 0 || level == literallevel) {

               for (int j = 0; j < l->m_arguments.size();  j++) {
                   types.insert(worldstate->getType(l->m_arguments[j].strid()));
                   if (newatomobjects.find(l->m_arguments[j].strid()) == newatomobjects.end()) {
                   newatomobjects.insert(l->m_arguments[j].strid());
                  // std::wcout << "mainworld.insert = " << l->m_arguments[j].strid().str() << std::endl;
                   }
               }
               if (alreadyadded.find(l) == alreadyadded.end()) {
                  alreadyadded.insert(l);
                  newworldstate->addAtom(l);
               }
               continue;
           }
       }
       currentatomobjects = newatomobjects;
    }


    for (int i = 0; i < worldstate->getFluentCount(); i++) {
        auto fl = worldstate->getFluent(i);
       Planner::Literal* l = std::get<0>(fl).get();
       int literallevel = getLevelFromLiteral(l);
       if ((literallevel & SEEALWAYS) != 0 || level == literallevel) {

                 newworldstate->addFluent(l, std::get<1>(fl));


       };
   }
}


void RTSHierarchicalControl::createUsedAtomsForSubproblem(Planner::IndependentWorldState* worldstate, const CUtiIDString& metanode,const CUtiIDString& objectid,  Planner::IndependentWorldState* newworldstate)
{
    //allowlist aus aktionen die ausgefuhrt werden mit level 1 (alle die das beruhren)
    //worldstate->getAtomsRecursively(metanode, 4,allowlist, atoms);
    //todo: bestimme delta in anderem state und update danach

    CUtiIDString metanode_type = worldstate->getType(metanode);


    int operationlevel = getLevelFromObjectSorts(metanode_type);

    std::set<CUtiIDString> types;
    std::set<CUtiIDString> currentatomobjects;
    currentatomobjects.insert(metanode);
    currentatomobjects.insert(objectid);
   // std::wcout << "currentatomobjects.insert = " << metanode.str() << " and objectid = "<< objectid.str()<< std::endl;
    std::set<CUtiIDString> newatomobjects;
    std::set<const Planner::Literal*> alreadyadded;
    CUtiIDString id_str =IDSTR("ID");
    std::set<CUtiIDString> atomtypes;





    //TODO: Unused objects predicates? -> unused objects -> function isLevelIndependent
    for (int step = 0; step < 4; step++)  {
        newatomobjects = currentatomobjects;
        for (int i = 0; i < worldstate->getStoredLiteralCount(); i++) {
           const Planner::Literal* l = &worldstate->getStoredLiteral(i);
           int literallevel = getLevelFromLiteral(l);
           if ((literallevel & SEEALWAYS) != 0) {

               for (int j = 0; j < l->m_arguments.size();  j++) {
                   types.insert(worldstate->getType(l->m_arguments[j].strid()));
                   if (newatomobjects.find(l->m_arguments[j].strid()) == newatomobjects.end()) {
                   newatomobjects.insert(l->m_arguments[j].strid());
                  // std::wcout << "currentatomobjects.insert = " << l->m_arguments[j].strid().str() << std::endl;
                   }
               }
               if (alreadyadded.find(l) == alreadyadded.end()) {
                  alreadyadded.insert(l);
                  newworldstate->addAtom(l);
               }
               continue;
           }
           if (literallevel != operationlevel+1) {
               if (l->hasObject(metanode) == false && l->hasObject(objectid) == false) {
               continue;
               }
           }
             if (l->hasObject(objectid)) {
//std::wcout << "l->hasObject(objectid) = " << l->toString() << std::endl;
             }
           /*if (l->m_name.str().endsWith(UTISTR("_parent"))) {
               continue;
           }*/
           if (l->m_arguments.size() == 0 || l->hasObjectOfType(currentatomobjects, worldstate, id_str, m_donotfollow))/* currentatomobjects.find(l->m_arguments[0].strid()) != currentatomobjects.end()*/ {
              // std::wcout << "atom_type = " << l->toString() << std::endl;
              atomtypes.insert(l->m_name);
              for (int j = 0; j < l->m_arguments.size();  j++) {
                  types.insert(worldstate->getType(l->m_arguments[j].strid()));
                  if (newatomobjects.find(l->m_arguments[j].strid()) == newatomobjects.end()) {
                  newatomobjects.insert(l->m_arguments[j].strid());
                 // std::wcout << "currentatomobjects.insert = " << l->m_arguments[j].strid().str() << std::endl;
                  }
              }
              if (alreadyadded.find(l) == alreadyadded.end()) {
                 alreadyadded.insert(l);
                 newworldstate->addAtom(l);
              }
           }
       }
       currentatomobjects = newatomobjects;
    }


    for (int i = 0; i < worldstate->getFluentCount(); i++) {
        auto fl = worldstate->getFluent(i);
       Planner::Literal* l = std::get<0>(fl).get();
       int literallevel = getLevelFromLiteral(l);
       if (literallevel != operationlevel+1 && (literallevel & SEEALWAYS) == 0) continue;

       if (l->hasObject(currentatomobjects) || (literallevel & SEEALWAYS) != 0) {
             newworldstate->addFluent(l, std::get<1>(fl));

       }
   }


}
