#ifndef HIERARCHICALCONTROL_H
#define HIERARCHICALCONTROL_H

#include "baseplaner.h"
#include "planninginstances.h"

struct TRTSEntityDesc {
    CUtiIDString m_name;
    int m_layer;
    std::vector<std::pair<CUtiIDString, int> > m_subentitycount;
    std::set<CUtiIDString> m_connectedstreets;
};

struct TRTSNPCDesc{

    CUtiIDString m_name;
    int m_layer;
    std::vector<std::pair<CUtiIDString, int> > m_subentitycount;
    inline int getTotalSubEntityCount()
    {
        int c = 0;
        for (auto& el: m_subentitycount)  {c += std::get<1>(el);}

        return c;
    }
};
const int SEEALWAYS = 64;



class RTSHierarchicalControl : public Planner::HierarchicalControlManager
{
protected:

    std::vector<CUtiIDString> m_materials;
    std::map<CUtiIDString, TRTSEntityDesc> m_entitydesc;
    std::map<CUtiIDString, TRTSNPCDesc> m_npcdesc;
    std::map<CUtiIDString, int> m_object_to_level;
    std::map<CUtiIDString, int> m_literal_name_to_level;
    std::set<CUtiIDString> m_donotfollow;


    int getLevelFromObjectSorts(const CUtiIDString& type);

    int getLevelFromLiteral(const Planner::Literal *l);
    CUtiIDString getMultiParam(const std::map<CUtiIDString, CUtiIDString>& params, const std::vector<CUtiIDString> &list);
    CUtiIDString getMultiParamByType(Planner::Task *t, const std::map<CUtiIDString, CUtiIDString>& params, const std::vector<CUtiIDString>& types, const std::vector<CUtiIDString> &list);
    bool updateInventory(Planner::IndependentWorldState* worldstate, const CUtiIDString& id,  const std::vector<CUtiIDString> &children, int layer);
    void handleSubNPCS(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, std::set<CUtiIDString>& stickynodes, int layer);
    void handleSubBuildings(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, int layer, std::vector<std::pair<CUtiIDString, int> > subentitycount,std::set<CUtiIDString> connectedstreets, std::set<CUtiIDString>& stickynodes);
public:
    RTSHierarchicalControl();
    virtual void updateNode(Planner::IndependentWorldState* worldstate, const CUtiIDString& id, int neighbourhooddistance, std::set<CUtiIDString> &stickynodes);
   //todo: repair npcs or buildings
    virtual std::tuple<bool, CUtiIDString, Planner::LiftedStateHeuristicSystem*> createFromEffectsGoal(Planner::Task* t, const CUtiIDString& taskname, Planner::IndependentWorldState* worldstate, const std::map<CUtiIDString, CUtiIDString> &params, const CUtiIDString& actionname, const CReferenceList<Planner::Literal>& effects, CStrings& goals, CStrings& newactions, CReferenceList<Planner::PlayerHierarchicalData>& hierarchicaldata);
    virtual void createMainProblem(Planner::IndependentWorldState* worldstate, int level, Planner::IndependentWorldState* newworldstate);
    virtual void createUsedAtomsForSubproblem(Planner::IndependentWorldState* worldstate, const CUtiIDString& metanode, const CUtiIDString& objectid, Planner::IndependentWorldState* newworldstate);


};

#endif // HIERARCHICALCONTROL_H
